/****************************************************************************
* VisualBoyAdvance 1.7.2
* Nintendo GameCube Wrapper
****************************************************************************/
#include <gccore.h>
#include <ogcsys.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <debug.h>
#include "mixer.h"
#include "gcpad.h"
#include "vmmem.h"
#include "pal60.h"


extern "C"
  {
#include "gx_supp.h"
#include "tbtime.h"
  }

/** VBA **/
#include "GBA.h"
#include "agbprint.h"
#include "Flash.h"
#include "Port.h"
#include "RTC.h"
#include "Sound.h"
#include "Text.h"
#include "unzip.h"
#include "Util.h"
#include "gb/GB.h"
#include "gb/gbGlobals.h"

/**
 * Globals
 */
int RGB_LOW_BITS_MASK=0x821;
int systemSaveUpdateCounter = SYSTEM_SAVE_NOT_UPDATED;
u16 systemGbPalette[24];
int systemDebug = 0;
int emulating = 0;
int systemFrameSkip = 0;
int systemRedShift = 0;
int systemBlueShift = 0;
int systemGreenShift = 0;
int systemColorDepth = 0;
int sensorX = 2047;
int sensorY = 2047;
u16 systemColorMap16[0x10000];
//u32 systemColorMap32[0x10000];
u32 *systemColorMap32 = (u32 *)&systemColorMap16;
bool systemSoundOn = false;
int systemVerbose = 0;
int cartridgeType = 0;
int srcWidth = 0;
int srcHeight = 0;
int destWidth = 0;
int destHeight = 0;
int srcPitch = 0;

#define WITHGX	1
#define DEBUGON 0

#if DEBUGON
const char *dbg_local_ip = "192.168.1.32";
const char *dbg_netmask = "255.255.255.0";
const char *dbg_gw = "192.168.1.100";
#endif

/*** 2D Video Globals ***/
GXRModeObj *vmode;		/*** Graphics Mode Object ***/
u32 *xfb[2] = { NULL, NULL };	/*** Framebuffers ***/
int whichfb = 0;		/*** Frame buffer toggle ***/

void debuggerOutput(char *, u32)
{}

void (*dbgOutput)(char *, u32) = debuggerOutput;

/**
 * Locals
 */
extern "C"
  {
    void SDInit( void );
    int gen_getdir( char *whichdir );
  };
extern char *direntries[1000];
extern char *MENU_GetLoadFile( char *whichdir );

struct EmulatedSystem emulator =
    {
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      false,
      0
    };

static tb_t start, now;
static u8 soundbuffer[3200] ATTRIBUTE_ALIGN(32);

extern u32 loadtimeradjust;

/****************************************************************************
* Initialise Video
*
* Before doing anything in libogc, it's recommended to configure a video
* output.
****************************************************************************/
static void
Initialise (void)
{
  VIDEO_Init ();		/*** ALWAYS CALL FIRST IN ANY LIBOGC PROJECT!
  				     Not only does it initialise the video 
  				     subsystem, but also sets up the ogc os
  				***/
#if DEBUGON
  DEBUG_Init(2424);
  _break();
#endif

  PAD_Init ();			/*** Initialise pads for input ***/

  vmode = &TVNtsc480IntDf;

  /*** Let libogc configure the mode ***/
  VIDEO_Configure (vmode);

  /*** Now configure the framebuffer.
       Really a framebuffer is just a chunk of memory
       to hold the display line by line.
  ***/

  xfb[0] = (u32 *) MEM_K0_TO_K1 (SYS_AllocateFramebuffer (vmode));
  /*** I prefer also to have a second buffer for double-buffering.
       This is not needed for the console demo.
  ***/
  xfb[1] = (u32 *) MEM_K0_TO_K1 (SYS_AllocateFramebuffer (vmode));

  /*** Define a console ***/
  console_init (xfb[0], 20, 64, vmode->fbWidth, vmode->xfbHeight,
                vmode->fbWidth * 2);

  /*** Clear framebuffer to black ***/
  VIDEO_ClearFrameBuffer (vmode, xfb[0], COLOR_BLACK);
  VIDEO_ClearFrameBuffer (vmode, xfb[1], COLOR_BLACK);

  /*** Set the framebuffer to be displayed at next VBlank ***/
  VIDEO_SetNextFramebuffer (xfb[0]);

  /*** Get the PAD status updated by libogc ***/
  VIDEO_SetPreRetraceCallback (PAD_ScanPads);
  VIDEO_SetBlack (0);

  int i;
  u32 *vreg = (u32 *)0xCC002000;
  for ( i = 0; i < 64; i++ )
	vreg[i] = vpal60[i];

  /*** Update the video for next vblank ***/
  VIDEO_Flush ();

  VIDEO_WaitVSync ();		/*** Wait for VBL ***/
  if (vmode->viTVMode & VI_NON_INTERLACE)
    VIDEO_WaitVSync ();

}

void systemMessage(int num, const char *msg, ...)
{
  /*** For now ... do nothing ***/
}
void systemFrame()
{}

void systemScreenCapture(int a)
{}

void systemShowSpeed(int speed)
{}

void system10Frames(int rate)
{
  static u32 autoFrameSkipLastTime = 0;
  static int frameskipadjust = 0;

  if ( cartridgeType == 1 )
    return;

  u32 time = systemGetClock();
  u32 diff = time - autoFrameSkipLastTime;
  int speed = 100;

  if(diff)
    speed = (1000000/rate)/diff;

  if(speed >= 98)
    {
      frameskipadjust++;

      if(frameskipadjust >= 3)
        {
          frameskipadjust=0;
          if(systemFrameSkip > 0)
            systemFrameSkip--;
        }
    }
  else
    {
      if(speed  < 80)
        frameskipadjust -= (90 - speed)/5;
      else if(systemFrameSkip < 9)
        frameskipadjust--;

      if(frameskipadjust <= -2)
        {
          frameskipadjust += 2;
          if(systemFrameSkip < 9)
            systemFrameSkip++;
        }
    }
  autoFrameSkipLastTime = time;

}

void systemUpdateMotionSensor()
{}

void systemGbBorderOn()
{}

void systemWriteDataToSoundBuffer()
{
  MIXER_AddSamples((u8 *)soundFinalWave, (cartridgeType == 1));
}

void systemSoundPause()
{}

void systemSoundResume()
{}

void systemSoundReset()
{}

void systemGbPrint(u8 *data,int pages,int feed,int palette, int contrast)
{}

void systemSoundShutdown()
{}

static void AudioPlayer( void )
{
  AUDIO_StopDMA();
  MIXER_GetSamples(soundbuffer, 3200);
  DCFlushRange(soundbuffer,3200);
  AUDIO_InitDMA((u32)soundbuffer,3200);
  AUDIO_StartDMA();
}

bool systemSoundInit()
{
  AUDIO_Init(NULL);
  AUDIO_SetDSPSampleRate(AI_SAMPLERATE_48KHZ);
  AUDIO_RegisterDMACallback(AudioPlayer);
  memset(soundbuffer, 0, 3200);

  printf("Audio Inited\n");
  return true;
}

bool systemPauseOnFrame()
{
  return false;
}

int systemGetSensorX()
{
  return sensorX;
}

int systemGetSensorY()
{
  return sensorY;
}

bool systemCanChangeSoundQuality()
{
  return true;
}

/****************************************************************************
* systemReadJoypads
****************************************************************************/
bool systemReadJoypads()
{
  return true;
}

u32 systemReadJoypad(int which)
{
  return NGCPad();
}

/****************************************************************************
* systemGetClock
*
* Returns number of milliseconds since program start
****************************************************************************/
u32 systemGetClock( void )
{
  mftb(&now);
  return tb_diff_msec(&now, &start) - loadtimeradjust;
}

/****************************************************************************
* systemDrawScreen
****************************************************************************/
void systemDrawScreen()
{
  /** GB / GBC Have oodles of time - so sync on VSync **/
#if WITHGX
  GX_Render( srcWidth, srcHeight, pix, srcPitch );
#endif
  if ( cartridgeType == 1 )
    {
      VIDEO_WaitVSync();
    }
}

/****************************************************************************
* Showdir
****************************************************************************/
int ShowDir( char *whichdir )
{
	int count;
	int i;
	count = gen_getdir( whichdir );
	printf("Found %d files\n",count);

	for ( i = 0; i < count; i++ )
		printf("%s\n", direntries[i]);
	while(1);
}

/****************************************************************************
* main
*
* Program entry
****************************************************************************/
int main()
{
  int i;
  //char *filename = "/SPEEDB.GB";
  //char *filename = "/VBA172/ROMS/SPACE.GBC";
  //char *filename = "/GALAGA.GBC";
  //char *filename = "/VBA172/ROMS/GALAXIAN.GB";
  //char *filename = "/VBA172/ROMS/MOTO.GBA";
  //char *filename = "/VBA172/ROMS/FF5.GBA";
  //char *filename = "/VBA172/ROMS/SIMS.GBA";
  //char *filename = "/VBA172/ROMS/SHREK.GBA";
  char filename[1024];
  char *fname;
  int vAspect = 0;
  int hAspect = 0;

  Initialise();

  printf("VisualBoyAdvance 1.7.2\n");
  printf("Nintendo GameCube Port\n");

  /** Kick off SD Lib **/
  SDInit();
  //ShowDir( "/VBA172/ROMS" );
  fname = MENU_GetLoadFile( "/VBA172/ROMS" );
  sprintf(filename, "/VBA172/ROMS/%s", fname);

  /** Build GBPalette **/
  for( i = 0; i < 24; )
    {
      systemGbPalette[i++] = (0x1f) | (0x1f << 5) | (0x1f << 10);
      systemGbPalette[i++] = (0x15) | (0x15 << 5) | (0x15 << 10);
      systemGbPalette[i++] = (0x0c) | (0x0c << 5) | (0x0c << 10);
      systemGbPalette[i++] = 0;
    }

  IMAGE_TYPE type = utilFindType(filename);

  switch( type )
    {
    case IMAGE_GBA:
      printf("GameBoy Advance Image\n");
      cartridgeType = 2;
      emulator = GBASystem;
      srcWidth = 240;
      srcHeight = 160;
      VMCPULoadROM(filename);
      /* Actual Visual Aspect is 1.57 */
      hAspect = 70;
      vAspect = 46;
      srcPitch = 484;
      break;

    case IMAGE_GB:
      printf("GameBoy Image\n");
      cartridgeType = 1;
      emulator = GBSystem;
      srcWidth = 160;
      srcHeight = 144;
      gbLoadRom(filename);
      /* Actual physical aspect is 1.0 */
      hAspect = 60;
      vAspect = 46;
      srcPitch = 324;
      break;

    default:
      printf("Unknown Image\n");
      while(1);
      break;

    }

  /** Set defaults **/
  flashSetSize(0x10000);
  rtcEnable(true);
  agbPrintEnable(false);

  /** Set palette etc - Fixed to RGB565 **/
  systemColorDepth = 16;
  systemRedShift = 11;
  systemGreenShift = 6;
  systemBlueShift = 0;

#if 0
  if ( cartridgeType == 2 )
    {
      for(i = 0; i < 0x10000; i++)
        {
          systemColorMap16[i] = (((i >> 1) & 0x1f) << systemBlueShift) |
                                (((i & 0x7c0) >> 6) << systemGreenShift) |
                                (((i & 0xf800) >> 11) << systemRedShift);
        }
    }
  else
#endif
    {
      for(i = 0; i < 0x10000; i++)
        {
          systemColorMap16[i] = ((i & 0x1f) << systemRedShift) |
                                (((i & 0x3e0) >> 5) << systemGreenShift) |
                                (((i & 0x7c00) >> 10) << systemBlueShift);
        }
    }

#if WITHGX
  /** Set GX **/
  GX_Start( srcWidth, srcHeight, hAspect, vAspect );
#endif

  /** Do sound **/
  soundOffFlag = false;
#if 0
  soundEnable(0x30f);
  soundDisable(~0x30f);
#endif
  //soundLowPass =  true;

  if ( cartridgeType == 1 )
    {
      soundQuality = 1;
      soundBufferLen = 1470 * 2;
      gbSoundReset();
      gbSoundSetQuality(soundQuality);
    }
  else
    {
      soundQuality = 2;
      soundBufferLen = 1470;
      soundSetQuality(soundQuality);
      cpuSaveType = 0;
      CPUInit("/VBA172/BIOS/BIOS.GBA", 1);
      CPUReset();
    }

  soundVolume = 0;
  systemSoundOn = true;

  soundInit();
  printf("Ready to rumble!\n");
  AudioPlayer();

  emulating = 1;
  /** Start system clock **/
  mftb(&start);

  while ( emulating )
    {
      emulator.emuMain(emulator.emuCount);
    }

  printf("Emulation complete\n");

  /** Never leave **/
  while(1);
  return 0;

}

